import 'dart:async';
import 'dart:math';

import 'package:flutter/cupertino.dart';
import 'package:flutter/gestures.dart';
import 'package:flutter_webview_desktop/src/desktop_webview_enums.dart';
import 'package:webview_flutter_platform_interface/webview_flutter_platform_interface.dart';
import 'desktop_webview_platform_controller.dart';

typedef DesktopWebviewWidgetFinishCallBack = void Function(DesktopWebviewController controller);

class DesktopWebviewWidget extends StatefulWidget {
  const DesktopWebviewWidget({
    Key? key,
    required this.width,
    required this.height,
    required this.initialUrl,
    required this.onFinishedCreated,
    required this.devicePixelRatio,
    required this.javascriptChannelRegistry,
    required this.webViewPlatformCallbacksHandler,
  }) : super(key: key);
  final double width;
  final double height;
  final double devicePixelRatio;
  final String initialUrl;
  final DesktopWebviewWidgetFinishCallBack onFinishedCreated;
  final JavascriptChannelRegistry javascriptChannelRegistry;
  final WebViewPlatformCallbacksHandler webViewPlatformCallbacksHandler;

  @override
  State<DesktopWebviewWidget> createState() => _DesktopWebviewWidgetState();
}

class _DesktopWebviewWidgetState extends State<DesktopWebviewWidget> {
  final _downButtons = <int, DesktopMousePointerButton>{};

  late DesktopWebviewController _controller;

  final MouseCursor _cursor = SystemMouseCursors.basic;
  StreamSubscription? _cursorSubscription;
  StreamSubscription? _goFocusSubscription;

  void _viewSize(BoxConstraints constraints) async {
    unawaited(
      _controller.setWebviewSize(
        Size(
          min(MediaQuery.of(context).size.width, constraints.maxWidth),
          min(MediaQuery.of(context).size.width, constraints.maxHeight),
        ),
        MediaQuery.of(context).devicePixelRatio,
      ),
    );
    // }
  }

  @override
  void initState() {
    super.initState();
    _controller = DesktopWebviewController(widget.javascriptChannelRegistry);

    _controller
        .initialize(
      widget.width.toInt(),
      widget.height.toInt(),
      widget.initialUrl,
      widget.devicePixelRatio,
    )
        .then((value) {
      widget.onFinishedCreated(_controller);
      _controller.setWindowOpenPolicy(DesktopWebviewClientWindowOpenPolicy.inSameWindow);
      if (widget.javascriptChannelRegistry.channels.isNotEmpty) {
        _controller.addJavascriptChannels((widget.javascriptChannelRegistry.channels.values.map((e) => e.name).toSet()));
      }
      setState(() {});
    });
    _controller.webview.loadingState.listen((event) {
      if (event == DesktopWebviewLoadState.completed || event == DesktopWebviewLoadState.loading) {
        widget.webViewPlatformCallbacksHandler.onPageFinished(widget.initialUrl);
      } else if (event == DesktopWebviewLoadState.start || event == DesktopWebviewLoadState.none) {
        widget.webViewPlatformCallbacksHandler.onPageStarted(widget.initialUrl);
      }
    });
  }

  @override
  void dispose() {
    _cursorSubscription?.cancel();
    _goFocusSubscription?.cancel();
    _controller.dispose();
    super.dispose();
  }

  @override
  void didUpdateWidget(covariant DesktopWebviewWidget oldWidget) {
    _controller.updateJavascriptChannels(widget.javascriptChannelRegistry.channels.values.toSet());
    super.didUpdateWidget(oldWidget);
  }

  @override
  Widget build(BuildContext context) {
    return FutureBuilder<int>(
        future: _controller.textureId(),
        builder: (context, data) {
          if (data.hasData && data.data != -1) {
            return _buildWebviewWidget(context, data.data ?? -1);
          }
          return const Text('_textureId is null ');
        });
  }

  Widget _buildWebviewWidget(BuildContext buildContext, int textureId) {
    return LayoutBuilder(
      builder: (context, constraints) {
        _viewSize(constraints);

        return Listener(
          onPointerHover: (ev) {
            // _controller._setCursorOffset(ev.localPosition);
          },
          onPointerDown: (ev) {
            final button = getDesktopButton(ev.buttons);
            _downButtons[ev.pointer] = button;
            _controller.setMouseClick(button, true, ev.localPosition);
          },
          onPointerUp: (ev) {
            final button = _downButtons.remove(ev.pointer);
            if (button != null) {
              _controller.setMouseClick(button, false, ev.localPosition);
            }
          },
          onPointerCancel: (ev) {
            final button = _downButtons.remove(ev.pointer);
            if (button != null) {
              _controller.setMouseClick(button, false, ev.localPosition);
            }
          },
          onPointerMove: (ev) {
            // if(ev.down)
            // print('movessss $ev ${ev.runtimeType}');
            _controller.setMouseMove(ev.localPosition, ev.delta);
          },
          onPointerSignal: (signal) {
            if (signal is PointerScrollEvent) {
              _controller.scrollDelta(signal.localPosition, signal.scrollDelta);
            }
          },
          child: MouseRegion(
            cursor: _cursor,
            child: Texture(textureId: textureId),
          ),
        );
      },
    );
  }
}
